% Copyright 2008 INRIA, F. Fages & J. Martin
% License GPL
%
% file ocg.pl
%

:- ensure_loaded(library(clpfd)).

% the rewriting symbol in geosr
:- op(1200, xfx, --->).

language_class(clpfd_sp, [clpfdsp, clpfdguisp]).
language_class(clpfd_cj, [clpfdcj]).

belongs_to_lang_class(Lang, LangClass) :-
	language_class(LangClass, Languages),
	memberchk(Lang, Languages).

object_language_class(ObjLangClass) :-
	object_language(ObjectLang),
	belongs_to_lang_class(ObjectLang, ObjLangClass).

% terms sets
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- dynamic cstr_terms/1,
	   domain_terms/1,
	   stree_terms/1,
	   label_terms/1,
	   optim_terms/1,
	   dynamic_terms/1,
	   varorder_terms/1,
	   valorder_terms/1,
	   dynvarorder_terms/1,
	   dynvalorder_terms/1,
	   gui_varname_terms/1,
	   gui_control_terms/1.

:- dynamic first_labeling/1.
:- dynamic cj_solver_cstr/1,
	   cj_solver_goal/1.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% query object code generation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
gen_objcode(1, []) :-
	!.
gen_objcode(0, []) :-
	!.
gen_objcode(dynamic(1), []) :-
	specialize_statements,
	!.
gen_objcode(dynamic(0), []) :-
	specialize_statements,
	!.

% clpfd SICStus
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
gen_objcode(IntermediateCode, [DecisionVariables, CstrGeneratedCode]) :-
	object_language(clpfdsp),
	!,		
	init_objcodegenerator,
	partition_term(IntermediateCode),	
	write('\"------------ Query\"'),nl,
	cstr_terms(Constraints),	
	generate_sp_code(Constraints, CstrGenCode, CstrReifCodeList),
	write('\"------------ Decision variables\"'),nl,
	label_terms(Labelings),
	domain_terms(DomainConstraints),
	variables_set([Labelings, DomainConstraints, Constraints], DecVariables),
	generate_sp_code(DecVariables, DecisionVariables, _),	
	write('\"------------ Statements\"'),nl,
	generate_sp_statements,
	list_to_set(CstrReifCodeList, CstrReifCode),
	append(CstrReifCode, CstrGenCode, CstrGeneratedCode).

% clpfd CHOCO
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
gen_objcode(IntermediateCode, [DecisionVariables, CstrGeneratedCode]) :-
	object_language(clpfdcj),
	!,		
	init_objcodegenerator,
	partition_term(IntermediateCode),	
	write('\"------------ Query\"'),nl,	
	cstr_terms(Constraints),
	generate_cj_code(Constraints, CstrGenCode, CstrReifCodeList),
	label_terms(Labelings),
	domain_terms(DomainConstraints),
	variables_set([Labelings, DomainConstraints, Constraints], DecVariables),
	generate_cj_code(DecVariables, DecisionVariables, _),
				%write('\"------------ Statements\"'),nl,
				%specialize_statements,	
	list_to_set(CstrReifCodeList, CstrReifCode0),
	cj_var_decl_first(CstrReifCode0, CstrReifCode),
	append(CstrReifCode, CstrGenCode, CstrGeneratedCode).

% clpfd + clpgui SICStus
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
gen_objcode(IntermediateCode, [GuiVarNamesGenCode,
			       GuiControlsGenCode,
			       ConstraintsGenCode]) :-
	object_language(clpfdguisp),	
	init_objcodegenerator,	
	partition_term(IntermediateCode),	
	domain_terms(DomainConstraints),
	variables_set(DomainConstraints, DomainVariables),	
	generate_sp_code(DomainVariables, DecisionVariables, _),
	cstr_terms(Constraints),
	generate_sp_code(Constraints, CstrGenCode, CstrReifCode),
	gui_varname_terms(GuiVarNames),
	gui_control_terms(GuiControls),					
	generate_sp_code(GuiVarNames, GuiVarNamesGenCode, _),
	gen_objcode_controls(GuiControls,
			     GuiControlsGenCode,
			     GuiControlsReifCode),	
	append(CstrReifCode, GuiControlsReifCode, CRC),
	append(CRC, CstrGenCode, ConstraintsGenCode).

gen_objcode_controls([], [], []).
gen_objcode_controls([GuiControl | GuiControls],
		     [GeneratedCode | GeneratedCodes], ReifCode) :-
	GuiControl = gui_control(button, Name, Term),	
	retract_terms([label_terms, stree_terms, optim_terms, cstr_terms]),	
	partition_term(Term),	
	generate_sp_code(gui_control(button, Name, IntCodeSearchList),
			GeneratedCode, RCode),	
	gen_objcode_controls(GuiControls, GeneratedCodes, RCodes),	
	append(RCodes, RCode, ReifCode).

init_objcodegenerator :-
	set_cg_mode(static),
	counter_init(gen_local_var_names),
	assert(first_labeling(true)),
	retractall(cj_solver_cstr(_)),
	assert(cj_solver_cstr([])),
	retractall(cj_solver_goal(_)),
	assert(cj_solver_goal([])),
	reset_dynamic_terms([cstr_terms, domain_terms,
			     label_terms, stree_terms, optim_terms,
			     varorder_terms, valorder_terms,
			     dynvarorder_terms, dynvalorder_terms,
			     gui_varname_terms, gui_control_terms]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% terms partitioning 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% {Domain} 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
partition_term(Term) :-
	functor(Term, Functor, Arity),
	lang(domain_preds, Preds),
	memberchk(Functor/Arity, Preds),
	!,
	store_term(Term, cstr_terms),
	store_term(Term, domain_terms).

% {Labeling, Search, Optimization}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
partition_term(Term) :-
	functor(Term, Functor, Arity),
	lang(label_preds, Preds),
	memberchk(Functor/Arity, Preds),
	!,
	store_term(Term, cstr_terms),
	store_term(Term, label_terms).
partition_term(Term) :-
	functor(Term, Functor, Arity),
	lang(stree_preds, Preds),
	memberchk(Functor/Arity, Preds),
	!,
	store_term(Term, cstr_terms),
	store_term(Term, stree_terms).
partition_term(Term) :-
	functor(Term, Functor, Arity),
	lang(optim_preds, Preds),
	memberchk(Functor/Arity, Preds),
	!,
	store_term(Term, cstr_terms),
	store_term(Term, optim_terms).

% {Variable,Value, Conjunct,Disjunct}
% Heuristics
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
partition_term(Term) :-
	Term =.. [F, Crits],
	lang(varorder_preds, Preds),
	memberchk(F/1, Preds),
	!,
	store_term(varorder(Crits)).
partition_term(Term) :-
	Term =.. [F, Crits],
	lang(valorder_preds, Preds),
	memberchk(F/1, Preds),
	!,
	store_term(valorder(Crits)).

% {GuiVarName, GuiControl}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
partition_term(Term) :-
	object_language(clpfdguisp),
	functor(Term, Functor, Arity),
	lang(gui_varname_preds, Preds),
	memberchk(Functor/Arity, Preds),
	!,
	store_term(Term, gui_varname_terms).
partition_term(Term) :-
	object_language(clpfdguisp),
	functor(Term, Functor, Arity),
	lang(gui_control_preds, Preds),
	memberchk(Functor/Arity, Preds),
	!,
	store_term(Term, gui_control_terms).

% {Dynamic}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
partition_term(dynamic(Term)) :-
	object_language_class(clpfd_sp),
	!,				
	store_term(dynamic(Term), cstr_terms).

% {And/Or Tree Node} (skip it)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
partition_term(node(_L, P)) :-
	!,
	partition_term(P).

% {Composition}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
partition_term((PL and PR)) :-
	!,
	partition_term(PL),
	partition_term(PR).

% {Constraints}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
partition_term(1) :-
	!.
partition_term(Term) :-
	is_constraint(Term),
	!,
	store_term(Term, cstr_terms).
partition_term(Term) :-	
	is_write(Term),
	!,
	store_term(Term, cstr_terms).

% {Dynamic Terms} 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
partition_term(Term) :-
	object_language_class(clpfd_sp),
	cg_mode(dynamic),
	!,
	store_term(Term, cstr_terms).

% {NotAKnownWord}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
partition_term(Term) :-
	object_language(ObjLang),
	raise_exception(code_generation(not_accepted_by_language(ObjLang),
					 [Term], [])).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% code generation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- ensure_loaded(['ocg.sp.pl']).

:- ensure_loaded(['ocg.cj.pl']).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% "type" checking
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% is Term a domain ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_domain(Term) :-
	functor(Term, Functor, Arity),	
	lang(domain_preds, Preds),
	memberchk(Functor/Arity, Preds).

% is Term a global constraint ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_glb_cstr(Term) :-	
	functor(Term, Functor, Arity),	
	lang(glbcstr_preds, Preds),
	memberchk(Functor/Arity, Preds).

% is Term a write ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_write(Term) :-
	functor(Term, Functor, 1),
	memberchk(Functor, [write]).

% is Term a constraint ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_constraint(Term) :-
	is_var(Term),
	!.
is_constraint(Term) :-
	is_domain(Term),
	!.
is_constraint(Term) :-
	is_compop(Term),
	!.
is_constraint(Term) :-
	is_logop(Term),
	!.
is_constraint(Term) :-
	is_glb_cstr(Term),
	!.
is_constraint(Term) :-
	cg_mode(dynamic),
	is_quantifier(Term).

% is Term a logical quantifier ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_quantifier(Term) :-
	functor(Term, Functor, Arity),	
	lang(quant_preds, Preds),
	memberchk(Functor/Arity, Preds).

% is Term a higher-order function ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_higherorder(Term) :-
	functor(Term, Functor, Arity),	
	lang(hof_preds, Preds),
	memberchk(Functor/Arity, Preds).

% is Term a builtin expression ?
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
is_builtin_expr(Term) :-
	var(Term),
	!.
is_builtin_expr(Term) :-
	is_var(Term),
	!.
is_builtin_expr(Term) :-
	integer(Term),
	!.
is_builtin_expr(Term) :-
	is_builtinop(Term),
	Term =.. [_ | Args],
	is_builtin_expr_args(Args).

is_builtin_expr_args([]).
is_builtin_expr_args([H | T]) :-
	is_builtin_expr(H),
	is_builtin_expr_args(T).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% auxiliaries
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
delete_variables(List, ListWoV, LV) :-
	reverse(List, RevList),
	delete_variables_rec(RevList, RevListWoV, LV),
	reverse(RevListWoV, ListWoV).

delete_variables_rec([], [], novar).
delete_variables_rec([H | T], T1,  H) :-
	is_var_name(H),
	!,
	delete_variables_rec(T, T1, _).
delete_variables_rec([H | T], T1,  H) :-
	atom(H),
	atom_codes(H, [0'_ | _]),
	!,
	delete_variables_rec(T, T1, _).
delete_variables_rec([H | T], [H | T1],  LV) :-
	delete_variables_rec(T, T1, LV).

% explicit reification 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
reify_compop(Term, RTerm, Add) :- 
	compound(Term),
	is_compop(Term),
	!,
	Term =.. [Op | Args],
	reify_compop(Args, RArgs, As),
	dict_sp(Op, TOp),
	ReifTerm =.. [TOp | RArgs],	
	fresh_symbol('R', ReifVar),
	RTerm = ReifVar,
	dict_sp(equiv, EquivOp),
	AddTerm =.. [EquivOp, ReifVar, ReifTerm],
	append(As, [AddTerm], Add).
reify_compop([], [], []) :-
	!.
reify_compop([H | T], [RT | RTs], Add) :-	
	reify_compop(H, RT, A),
	reify_compop(T, RTs, As),
	!,
	append(A, As, Add).
reify_compop(Term, RTerm, Add) :-	
	generate_sp_code(Term, RTerm, Add).

contains_compop([H | _]) :-
	contains_compop(H),
	!.
contains_compop([_ | T]) :-
	contains_compop(T),
	!.
contains_compop(Term):-
	compound(Term),
	is_compop(Term).

% global constraint utilities
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% disjoint2  
%%%%%%%%%%%%%%%%%%%%
disjoint_data([], [], [], []).
disjoint_data([X | Xs], [Y | Ys], [S | Ss], [r(X, S, Y, S) | Rs]) :-
	disjoint_data(Xs, Ys, Ss, Rs).

% cumulative  
%%%%%%%%%%%%%%%%%%%%
mktasks([], [], [], [], []).
mktasks([O | Os], [D | Ds], [H | Hs], [E | Es], [task(O, D, E, H, 0) | Tasks]) :-
	mktasks(Os, Ds, Hs, Es, Tasks).


traverse_presbox_shapes([],[]).
traverse_presbox_shapes([PreSboxShape | PreSboxShapes], SboxesShapes) :-
	traverse_presbox_shapes(PreSboxShapes,SboxShapes),
	traverse_presbox_boxes(PreSboxShape, Sboxes),
	append(Sboxes, SboxShapes, SboxesShapes).

traverse_presbox_boxes([],[]).
traverse_presbox_boxes([PreSboxBox | PreSboxBoxes],[Sbox | Sboxes]) :-
	traverse_presbox_boxes(PreSboxBoxes,Sboxes),
	PreSboxBox = [Sid, Offset, Sizes],
	Sbox =.. [sbox, Sid, Offset, Sizes].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% includes
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- ensure_loaded([order]).
