% Copyright 2008 INRIA, F. Fages & J. Martin
% License GPL
%
% file subst.pl
%

:- ensure_loaded(library(lists)).
:- ensure_loaded(library(sets)).
:- use_module(library(avl)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% environment handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
env_new(NewEnv) :-
	var(NewEnv),
	empty_avl(NewEnv).

env_member(Expr, Env, Value) :-
	avl_fetch(Expr, Env, Value).

env_add(Expr, Env, Value, NewEnv) :-
	avl_store(Expr, Env, Value, NewEnv).

env_remove(Expr, Env, NewEnv) :-
	avl_delete(Expr, Env, _, NewEnv).

env_print(Env) :-
	portray_avl(Env),nl.

env_list(Env, List) :-
	avl_to_list(Env, List).

env_vars_list(Env, KeyList) :-
	env_list(Env, List),
	env_vars_list_rec(List, KeyList).

env_vars_list_rec([], []).
env_vars_list_rec([K-_ | T], [K | KT]) :-	
	env_vars_list_rec(T, KT).

% lists 
%%%%%%%%%%%%%%%%%%%%%
env_member_list([], _, [], []).
env_member_list([Expr | Exprs], Env, [Value | Values], [Expr | ExprMembers]) :-
	env_member(Expr, Env, Value), !,
	env_member_list(Exprs, Env, Values, ExprMembers).
env_member_list([_ | Exprs], Env, Values, ExprMembers) :-
	env_member_list(Exprs, Env, Values, ExprMembers).

env_add_list([], Env, [], Env).
env_add_list([Expr | Exprs], Env, [Value | Values], NewEnv) :-	
	env_add(Expr, Env, Value, NewEnv1),	
	env_add_list(Exprs, NewEnv1, Values, NewEnv).  

env_remove_list([], Env, Env).
env_remove_list([Expr | Exprs], Env, NewEnv) :-
	env_remove(Expr, Env, NewEnv1),
	env_remove_list(Exprs, NewEnv1, NewEnv).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% alpha-conversion
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
alpha_conversion(Var, List, Expr, AlphaVar-AlphaExpr) :-	
	Var = var(W, _),
	counter_inc(global_vars,N),
	AlphaVar = var(W, N),
	variables_set(List-Expr, Vars),
	env_new(EmptyEnv),
	env_add_list(Vars, EmptyEnv, Vars, Env1),
	env_add(Var, Env1, AlphaVar, Env),
	substitution(Env, Expr, '', _OutEnv, AlphaExpr).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% substitution 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Expr is a free variable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
substitution(InEnv, Expr, VarSrcTerm, OutEnv, SubstExpr) :-
	is_anonym_var(Expr),
	!,
	counter_inc(local_vars, I),
	SubstExpr = free_var(VarSrcTerm, 0, I),
	OutEnv = InEnv.

substitution(InEnv, free_var(X, N, I), VarSrcTerm, OutEnv, SubstExpr) :-
	!,	
	(
	  VarSrcTerm == ''
	->
	  SubstExpr = free_var(X, N, I)
	;
	  SubstExpr = free_var(VarSrcTerm, X, I)
	),
	env_add(free_var(X, N, I), InEnv, SubstExpr, OutEnv).

% Expr is a binded variable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
substitution(InEnv, Expr, _, OutEnv, SubstExpr) :- 
	is_var(Expr),
	!,
	(
	  env_member(Expr, InEnv, VarName)
	->	    
	  SubstExpr = VarName,
	  OutEnv = InEnv	
	;
	  error(variable_binding, "~w cannot be free", [Expr])
	).

% Expr is a list
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
substitution(InEnv, Expr, VarSrcTerm, OutEnv, SubstExpr) :-
	is_list(Expr),
	!,
	substitution_list(InEnv, Expr, VarSrcTerm, OutEnv, SubstExpr).

% Expr is a binder
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
substitution(InEnv, Expr, VarSrcTerm, OutEnv, SubstExpr) :-
	compound(Expr), 
	Expr =.. [F, var(V, N), L | Args],
	is_binder(F),
	!,
	last(Args, BindedExpr),				
	substitution(InEnv, L, VarSrcTerm, OutEnvL, SubstL),
	(
	  env_member(var(V, N), OutEnvL, _)
	->	
	  counter_inc(global_vars, I),
				% var(V, N-I)
	  SubstV = var(V, I),
	  env_add(var(V, N), OutEnvL, SubstV, InEnvBindedExpr),
	  substitution(InEnvBindedExpr, BindedExpr, '', OutEnvBinder, BindedExprSubst),
	  OutEnv = InEnv
	;
	  env_add(var(V,N), OutEnvL, var(V,N), InEnvBindedExpr),
	  substitution(InEnvBindedExpr, BindedExpr, VarSrcTerm, OutEnvBinder, BindedExprSubst),
	  env_remove(var(V,N), OutEnvBinder, OutEnv),
	  SubstV = var(V,N)
	),
	delete(Args, BindedExpr, Args1),
	substitution(OutEnv, Args1, VarSrcTerm, _, SubstArgs1),
	append(SubstArgs1, [BindedExprSubst], SubstArgs),
	SubstExpr =.. [F, SubstV, SubstL | SubstArgs].

% Expr is a conditional expression (which is a binder)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
substitution(InEnv, (BindedExpr if TExpr), VarSrcTerm, OutEnv, SubstExpr) :-
	!,
	variables_set(TExpr, TExprVars),
	env_add_list(TExprVars, InEnv, TExprVars, InEnvBindedExpr),
	substitution(InEnvBindedExpr, BindedExpr, VarSrcTerm, OutEnvBinder, SubstExpr),
	env_remove_list(TExprVars, OutEnvBinder, OutEnv).

% Expr is a compound
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
substitution(InEnv, Expr, VarSrcTerm, OutEnv, SubstExpr) :-
	compound(Expr),
	!,
	Expr =.. [F | Args],
				%write(compound(Expr)),nl,	
	substitution_list(InEnv, Args, VarSrcTerm, OutEnv, SubstArgs),
	SubstExpr =.. [F | SubstArgs].

% Expr is anything else
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
substitution(InEnv, Expr, _, InEnv, Expr).

% list subsitution
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
substitution_list(InEnv, [], _, InEnv, []).
substitution_list(InEnv, [Expr | Exprs], VarSrcTerm, OutEnv, [SubstExpr | SubstExprs]) :-
	substitution(InEnv, Expr, VarSrcTerm, OutEnvSoFar, SubstExpr),
	substitution_list(OutEnvSoFar, Exprs, VarSrcTerm, OutEnv, SubstExprs).

% Is Term closed w.r.t. BindedVars ? 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
closed_term(Term, BindedVars) :-
	variables(Term, TermVars),
	!,
	list_to_set(BindedVars, BindedVarsSet),
	list_to_set(TermVars, TermVarsSet),
	subset(TermVarsSet, BindedVarsSet).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% variables occurences 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% set
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
variables_set(Expr, Set) :-
	variables(Expr, Bag),
	list_to_set(Bag, Set).

% multiset
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
variables(Expr, Bag) :-
	variables_bag(Expr, Bag, []).

variables_bag(Expr) -->
	{
	 is_var(Expr)
	},
	!,
	[Expr]. 
variables_bag(Expr) -->
	{
	 simple(Expr)
	},
	!.
variables_bag(Expr) -->
	{
	 functor(Expr, _, K)
	},
	variables_bag(0, K, Expr).

variables_bag(K, K, _) -->
	!.
variables_bag(I, K, Expr) -->
	{
	 J is I+1,
	 arg(J, Expr, Arg)
	},
	variables_bag(Arg),
	variables_bag(J, K, Expr).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% free variables naming
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
name_free_variables(Term, BindedVars, NTerm) :-
	is_var(Term),
	!,
	(
	  memberchk(Term, BindedVars)
	->
	  NTerm = Term
	;
	  Term = var(X, N),
				% '' -became-> 0
	  NTerm = free_var(X, 0, N)
	).
name_free_variables(Term, _, Term) :-
	atomic(Term),
	!.
name_free_variables(Term, BindedVars, NTerm) :- 
	is_list(Term),
	!,
	name_free_variables_rec(Term, BindedVars, NTerm).
name_free_variables((Expr if CExpr), BindedVars, NTerm) :-
	!,
	variables_set(CExpr, CVars),
	append(BindedVars, CVars, NBindedVars),
	name_free_variables(CExpr, NBindedVars, NCExpr),
	name_free_variables(Expr, NBindedVars, NExpr),
	NTerm =.. (NExpr if NCExpr).
name_free_variables(Term, BindedVars, NTerm) :- 
	compound(Term),
	Term =.. [F, V, L | Args],
	is_binder(F),
	!,
	name_free_variables(L, BindedVars, NL),
	name_free_variables_rec(Args, [V | BindedVars], NArgs),
	NTerm =.. [F, V, NL | NArgs].
name_free_variables(Term, BindedVars, NTerm) :- 
	Term =.. [F | Args],
	name_free_variables_rec(Args, BindedVars, NArgs),
	NTerm =.. [F | NArgs].

name_free_variables_rec([], _, []).
name_free_variables_rec([T | Ts], BindedVars, [NT | NTs]) :-	
	name_free_variables(T, BindedVars, NT),
	name_free_variables_rec(Ts, BindedVars, NTs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% anonymous variables naming
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

name_anonymous_variables(Term) :-
	counter_init(local_vars),
	name_anonym_vars(Term).

name_anonym_vars(var('_', I)) :-
	!,
	counter_inc(local_vars, I).
name_anonym_vars(T) :-
	T =.. [_ | Args],
	name_anonym_vars_rec(Args).

name_anonym_vars_rec([]).
name_anonym_vars_rec([H | T]) :-
	name_anonym_vars(H),
	name_anonym_vars_rec(T).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% variables naming
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
name_variables(T, [], T).
name_variables(Tx, VarNames, T) :-
	build_varval_pairs(VarNamed, VarNames, Var),	
	copy_term(Var-Tx, VarNamed-T).

build_varval_pairs(var(V, 0), [(V = N)], N) :-
	!.
build_varval_pairs(var(V, 0)-U, [(V = N) | VarNames], N-M) :-
	build_varval_pairs(U, VarNames, M).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% (textual) substitution 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subst(Var, Val, Term, TermSubst) :-
	simple(Term),
	!,
	(
	  Var == Term
	->
	  TermSubst = Val
	;
	  TermSubst = Term
	).
subst(Var, Val, Term, TermSubst) :-
	functor(Term, F, K),
	functor(Var, F, K),
	!,
	(
	  Var == Term
	->
	  TermSubst = Val
	;
	  TermSubst = Term
	).
subst(Var, Val, Term, TermSubst) :-
	functor(Term, F, K),
	functor(TermSubst, F, K),
	subst(0, K, Var, Val, Term, TermSubst).

subst(K, K, _, _, _, _) :-
	!.
subst(I, K, Var, Val, Term, TermSubst) :-
	J is I+1,
	arg(J, Term, X),
	arg(J, TermSubst, Y),
	subst(Var, Val, X, Y),
	subst(J, K, Var, Val, Term, TermSubst).

%  substitute (, | ;) for (and | or)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subst_and_or(and, ',') :-
	!.
subst_and_or(or, ';') :-
	!.
subst_and_or(Term, Term) :-
	simple(Term),
	!.
subst_and_or(and(X1,Y1), (X2,Y2)) :-
	!,
	subst_and_or(X1, X2),
	subst_and_or(Y1, Y2).
subst_and_or(or(X1,Y1), (X2;Y2)) :-
	!,
	subst_and_or(X1, X2),
	subst_and_or(Y1, Y2).
subst_and_or(Term, TermSubst) :-
	functor(Term, F, K),
	functor(TermSubst, F, K),
	subst_and_or(0, K, Term, TermSubst).

subst_and_or(K, K, _, _) :-
	!.
subst_and_or(I, K, Term, TermSubst) :-
	J is I+1,
	arg(J, Term, X),
	arg(J, TermSubst, Y),
	subst_and_or(X, Y),
	subst_and_or(J, K, Term, TermSubst).


%  substitute free variables reference name term
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subst_free_var(Term, _Name, Term) :-
	simple(Term),
	!.
subst_free_var(free_var(X, I, N), SubstForTerm, free_var(SubstForTerm, I, N)) :-
	(
	  is_underscore(X)
	;
	  is_upper_case(X)
	),
	!.
subst_free_var(Term, SubstForTerm, TermSubst) :-
	functor(Term, F, K),
	functor(TermSubst, F, K),
	subst_free_var(0, K, Term, SubstForTerm, TermSubst).

subst_free_var(K, K, _, _, _) :-
	!.
subst_free_var(I, K, Term, SubstForTerm, TermSubst) :-
	J is I+1,
	arg(J, Term, X),
	arg(J, TermSubst, Y),
	subst_free_var(X, SubstForTerm, Y),
	subst_free_var(J, K, Term, SubstForTerm, TermSubst).



%  substitute 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Is Sub a subterm of Term ? (works with lists)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
subterm(Sub, Term) :-
	nonvar(Term), 
	Sub=Term.
subterm(Sub, Term) :-
	nonvar(Term),
	functor(Term, _, N),
	subterm(N, Sub, Term).

subterm(N, Sub, Term) :-
	N > 0,
	arg(N, Term, Arg),
	subterm(Sub, Arg).
subterm(N, Sub, Term) :-
	N > 1,
	N1 is N-1,
	subterm(N1, Sub, Term).
